/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.communication.connections
 * File:    SSRC.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.communication.connections;

import java.io.IOException;
import java.util.Timer;
import java.util.TimerTask;

import javax.xml.bind.JAXBException;

import org.apache.http.HttpException;
import org.apache.log4j.Logger;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType;
import org.w3._2003._05.soap_envelope.Body;
import org.w3._2003._05.soap_envelope.Envelope;
import org.w3._2003._05.soap_envelope.Header;

import de.fhhannover.inform.iron.mapserver.communication.Client;
import de.fhhannover.inform.iron.mapserver.communication.DeprecatedDataModelServiceAccessor;
import de.fhhannover.inform.iron.mapserver.datamodel.DataModelService;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoPollResultAvailableException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchPublisherException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.RequestCreationException;
import de.fhhannover.inform.iron.mapserver.logging.Logging;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.RequestFactory;
import de.fhhannover.inform.iron.mapserver.utils.MessageTransformer;

/**
 * SSRC class
 * 
 * Class for synchronous send-receive channel connections (SSRC)
 * 
 * @author Waldemar Bender
 * @version 0.1
 * 
 * created: ???
 *
 * changes:
 *  12.06.10 - make sure to the old timer before starting a new..
 *  14.02.2010 - comments, cleanup
 *
 */
public class SSRC {
	private String mPublisherID;
	private String mSessionID;
	private Timer mSessionTimer;
	private final int TIMEOUT = 180;
	private DataModelService dms;
	//private ArrayList<PollResult> resultList = new ArrayList<PollResult>();
	private boolean poll = false;
	private boolean pollResultAvailable = false;
	private Client arc;
	private static Logger logger;
	static {
		logger = Logging.getTheLogger();
	}
	
	public SSRC (String pubID, String sessID){
		mPublisherID = pubID;
		mSessionID = sessID;
		dms = DeprecatedDataModelServiceAccessor.getDataModelServiceInstance();
		
		
		// Why, we should only start a timer if we detect that the connection
		// was closed
		// startTimer(TIMEOUT);
	}



	public String getPublisherID() {
		return mPublisherID;
	}


	public String getSessionID() {
		return mSessionID;
	}
	
	public void renewSession(){
		stopTimer();
		// Why?
		// startTimer(TIMEOUT);
		logger.info("renewSession: "+ this.getSessionID());		
	}
	
	public void stopTimer() {
		if (mSessionTimer != null) {
			logger.debug("Stopping Timer for Publisher " + mPublisherID);
			mSessionTimer.cancel();
			mSessionTimer = null;
		}
	}
	
	public void startTimer (int timeInSecond){
		stopTimer();
		logger.trace("Starting Timer for Publisher " + mPublisherID);
		mSessionTimer = new Timer();
		mSessionTimer.schedule(new SessionTimer(this), (long) timeInSecond *1000);
	}
	
	public void poll (Client client) {
		arc = client;
		if (!pollResultAvailable()) {
			poll = true;
		} else {
			sendPollResponse();
		}
		
	}
	
	public void  sendPollResponse () {
		try {
			PollResult pr = DeprecatedDataModelServiceAccessor.getDataModelServiceInstance().getPollResultFor(mSessionID);
			ResponseType prt = MessageTransformer.getInstance().transformPollResult(pr);
			Envelope env = new Envelope();
			env.setHeader(new Header());
			Body body = new Body();
			body.setResponse(prt);
			env.setBody(body);
			logger.trace("Sending a poll result!!!");
			arc.sendResponse(env);
			logger.trace("Reseting poll state");
			poll = false;
			pollResultAvailable = false;
		} catch (HttpException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (JAXBException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NoPollResultAvailableException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NoSuchPublisherException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public boolean isWaitingForPollResult () {
		return poll;
	}
	
	public boolean pollResultAvailable () {
		return pollResultAvailable;
	}
	
	public void setPollResultAvailable() {
		logger.trace("Setting PollResultAvailable to true");
		pollResultAvailable = true;
	}
	
	public void closeConnection() {
		logger.info("Connection will be closed. " +
				"SessionID: " + this.getSessionID() + "PublisherID: " + this.getPublisherID());
		stopTimer();
		
		try {
			dms.endSession(RequestFactory.getInstance().createEndSessionRequest(mSessionID));
		} catch (RequestCreationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NoSuchPublisherException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} finally{
			try {
				if (arc != null) {
					arc.getSocket().close();					
				}
				ConnectionContainer.getInstance().closeSession(this);
				this.finalize();
			} catch (Throwable e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	
	public String toString() {
		int plen = (mPublisherID.length() < 10) ? mPublisherID.length() : 10;
		int slen = (mSessionID.length() < 10) ? mSessionID.length() : 10;
		StringBuilder sb = new StringBuilder();
		sb.append("PID=\"" + mPublisherID.substring(0, plen));
		sb.append("\" SID=\"" + mSessionID.substring(0, slen));
		if(mSessionTimer != null) {
			sb.append("\" Timer=on  ");
		} else {
			sb.append("\" Timer=off ");
		}
		
		if(isWaitingForPollResult()) {
			sb.append("poll pending ");
		} else {
			sb.append("no poll      ");
		}

		if(pollResultAvailable()) {
			sb.append("pollres avail ");
		} else {
			sb.append("no pollres    ");
		}
		return sb.toString();
	}
	
	class SessionTimer extends TimerTask{
		private SSRC mCon;
		public SessionTimer (SSRC con){
			super();
			mCon = con;
			
		}
		public void run(){
			logger.trace("TimerTask of publisher " + mCon.mPublisherID + " now ACTIVE!");
			mCon.closeConnection();
		}
	}
}
