/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.configuration.communication
 * File:    ConfigurationManager.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.configuration.communication;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Date;
import java.util.Random;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.configuration.ConfigurationReader;
import de.fhhannover.inform.iron.mapserver.logging.Logging;

/**
 * ConfigurationManager class
 * 
 * Manages the properties file which contains publisherID and hash values
 * 
 * @author wb, dl
 * @version 0.1
 * 
 * created: ???
 * 
 * changes:
 *  14.02.10 - added comments, added logger
 *
 */
public class ConfigurationManager {

	private static ConfigurationManager mInstance;
	private ConfigurationReader mProps;
	private Random mRandom = new Random();
	private long mSeed = new Date().getTime() -5000;
	private MessageDigest mDigest;
	private static Logger logger;
	static {
		logger = Logging.getTheLogger();
	}
	
	private ConfigurationManager(){
		if (mProps == null){
			mProps = ConfigurationReader.getInstance();
		}
		try {
			mDigest = java.security.MessageDigest.getInstance("MD5");
		} catch (NoSuchAlgorithmException e) {
			logger.error("MD5 ist not available!");
		}
	}
	
	public static ConfigurationManager getInstance(){
		if (mInstance == null){
			mInstance = new ConfigurationManager();
		}
		return mInstance;
	}

	
	public String getPublisherID(String hash){
		return checkPublisherID(hash);

	}
	
	/**
	 * Checks if hash is associated with publisherID
	 * 
	 * @param hash
	 * @param publisherID
	 * @return
	 */
	public boolean isPublisherID(String hash, String publisherID){
		String pubID;

		if ((pubID = mProps.getProperty(hash)) != null){
			if (pubID.equals(publisherID)){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Check if hash is in properties file, if not a new publisherID will be returned
	 * 
	 * @param hash
	 * @return
	 */
	private String checkPublisherID(String hash){
		String pubID;

		if ((pubID = mProps.getProperty(hash)) == null) {
			pubID = getNewPublisherID(hash);			
		}

		return pubID;
	}
	
	/**
	 * Generate publisherID
	 * 
	 * @param hash
	 * @return
	 */
	private String getNewPublisherID(String hash){
		long random;
		byte []randomHash;
		String pubID;
		int j = 0;
		do{
			j++;	
			mSeed += 23416;
			mRandom.setSeed(mSeed);
			mSeed -= 12311;
			random = mRandom.nextLong();
			mDigest.update(toByteArray(random));
			randomHash = mDigest.digest();
	    	StringBuffer strbuf = new StringBuffer();
	    	for (int i = 0; i < randomHash.length; i++) {
	  			strbuf.append(toHexString(randomHash[i]));
	    	}
	    	pubID = strbuf.toString();
	    	logger.trace("Aufruf " + j);
		} while(!setPublisherID(hash, pubID));
		
    	return pubID;
	}
	
	/**
	 * Add new hash with publisherID to properties file if file does not contain publisherID
	 * 
	 * @param hash
	 * @param pubID
	 * @return
	 */
	private boolean setPublisherID(String hash, String pubID){
		if (mProps.containsValue(pubID)){
			return false;
		}
		mProps.put(hash, pubID);

		return true;
	}
	
  	private String toHexString(byte b) {
   		int value = (b & 0x7F) + (b < 0 ? 128 : 0);
	    
    		String ret = (value < 16 ? "0" : "");
    		ret += Integer.toHexString(value).toUpperCase();
	    
		return ret;
  	}
  	
  	private byte[] toByteArray(long l) {
  		byte[] ba=new byte[8];
  		for(int i=0; i<8; i++) {
  			ba[i] = new Long((l>>(i<<3))&255L).byteValue();
  		}
  		return ba;
  	}
	
}
