/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    SubscriptionService.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.AlreadyObservedException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchPublisherException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchSubscriptionException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Link;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.clientmapping.Publisher;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaState;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.logging.Logging;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;
import de.fhhannover.inform.iron.mapserver.messages.SubPublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubSubscriptionRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubscriptionDelete;
import de.fhhannover.inform.iron.mapserver.messages.SubscriptionRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubscriptionUpdate;

/**
 * Class to handle SubscriptionRequests
 * 
 * @version 0.1
 * @author aw
 *
 */

/*
 * created: 30.04.10
 * 
 * changes:
 *  30.04.10 aw - created just a dummy class
 * 	28.05.10 vp - create startident if it's missing
 *	08.06.10 aw - some commenting and fixes
 *  12.06.10 aw - subscriptions with new notifyresult shall only notify the
 *   observer during updateSubscriptions
 * 
 */
public class SubscriptionService {
	
	private static Logger logger;
	
	static {
		logger = Logging.getTheLogger();
	}
	
	private PublisherRep publisherRep;
	private MetaController metaController;
	private IdentifierRep identifierRep;
	
	private SubscriptionObserver observer;
	
	private Set<Subscription> changedSubscriptions;

	SubscriptionService(IdentifierRep iRep, PublisherRep pRep, MetaController mCtl) {
		identifierRep	= iRep;
		publisherRep	= pRep;
		metaController	= mCtl;
		
		changedSubscriptions = new HashSet<Subscription>();
	}
	
	/**
	 * Handling of a {@link SubscriptionRequest} which is forwarded by the
	 * {@link DataModelService}.
	 * Look at all teh {@link SubPublishRequest} objects contained in the
	 * {@link SubscriptionRequest} and dispatch whether it's a update or
	 * a delete.
	 * 
	 * @param subscriptionReq
	 * @throws NoSuchPublisherException
	 * @throws NoSuchSubscriptionException
	 */
	void subscribe(SubscriptionRequest subscriptionReq) throws NoSuchPublisherException, NoSuchSubscriptionException {
		
		Publisher publisher = publisherRep.getPublisherBySessionId(subscriptionReq.getSessionId());
		
		for (SubSubscriptionRequest ssr : subscriptionReq.getSubSubscriptionRequests()) {
			if (ssr instanceof SubscriptionUpdate) {
				
					processSubscribeUpdate(publisher, (SubscriptionUpdate)ssr);
					
			} else if (ssr instanceof SubscriptionDelete) {
				
					processSubscribeDelete(publisher, (SubscriptionDelete)ssr);
					
			}
		}
	}

	/**
	 * Delete a {@link Subscription} from a {@link Publisher}
	 * 
	 * @param publisher
	 * @param ssr
	 * @throws NoSuchSubscriptionException
	 */
	private void processSubscribeDelete(Publisher publisher, SubscriptionDelete ssr)
													throws NoSuchSubscriptionException {
		Map<String, Subscription> subscriptions = publisher.getSubscriptions();
		
		Subscription sub = subscriptions.get(ssr.getName());
		
		if (sub != null) {
			logger.debug("Deleting subscription \"" + sub.getName() + "\" of " + publisher);
			clearContainers(sub);
			subscriptions.remove(ssr.getName());
		} else {
			throw new NoSuchSubscriptionException("Subscription with name " +
					ssr.getName() + " not available");
		}
	}

	

	
	
	/**
	 * Create a new {@link Subscription}, attach it to a {@link Publisher},
	 * run the first search.
	 * 
	 * If a {@link Subscription} with the same name is already there reset
	 * it and update the {@link SearchResult} objects
	 * 
	 * @param publisher
	 * @param ssr
	 */
	private void processSubscribeUpdate(Publisher publisher, SubscriptionUpdate ssr) {
		
		logger.debug("Updating subscription \"" + ssr.getName() + "\" of " + publisher);
		
		Subscription sub = publisher.getSubscriptions().get(ssr.getName());
	
		// subscription was already there
		if (sub != null) {
			logger.trace("Updating old subscription");
			sub.setSearchRequest(ssr.getSearchRequest());
			sub.resetResults();
			clearContainers(sub);
		} else {
			sub = new Subscription(ssr.getName(), ssr.getSearchRequest());
			publisher.getSubscriptions().put(ssr.getName(), sub);
		}
		
		// if a startident is missing in the map, it must be created
		identifierRep.getOrCreateIdentifier(ssr.getSearchRequest().getStartIdentifier());

		//
		// create objects for the first search in the graph
		// why do we use a SearchResultImpReference here???
		// and then just add all the metadata objects into the the real
		// SearchResult, which is??? Uhh...
		SearchResultReferenceImpl sRes = new SearchResultReferenceImpl(publisher.getSessionId());
		SearchResultBuilder builder = new SearchResultBuilderImpl(publisher.getSessionId(), sRes);
		Searcher searcher = new Searcher(identifierRep, ssr.getSearchRequest(), builder);
		searcher.runSearch();
		
		builder = new SearchResultBuilderImpl(publisher.getSessionId(), sub.getSearchResult());
		
		for (MetadataContainer mc : sRes.getContainers()) {
			mc.addSubscription(sub);
			sub.addContainer(mc);
			
			builder.addMetadataContainer(mc);
			
			for (Metadata m : sRes.getMetadataFor(mc))
				builder.addMetadata(mc, m);
			
		}

		// TODO: REREAD Specification if no metadata is available...
		// However, this SearchResult always contains the identifier where the
		// search was begun.
		
		if (!sub.getSearchResult().isEmpty()) {
			sub.setChanged();
			logger.debug("Informing Publisher " + publisher + " about it's first SearchResult");
			notifyChangedSubscriptions(publisher);
		}
	}
	

	void updateSubscriptions(Map<Integer, Metadata> changes) {
		logger.debug("Update Subscriptions...");
		logger.trace("Changed metadata:");
		for (Metadata m : changes.values()) {
			logger.trace("\t" + m.getType().getTypeString() + " == " + m.getState().toString());
//			logger.trace("\t\t Attached Subscriptions:");
//			for (Subscription sub : m.getContainerReference().getSubscriptions()) {
//				logger.trace("\t\t " + sub.getName());
//			}
		}
		commitChanges(changes);
		notifyChangedSubscriptions();
	}
	
	/**
	 * This is farked...
	 * 
	 * FIXME: That needs to be fixed, it's unreadable.. and a lot of
	 * code looks redundant...
	 * 
	 * @param changes
	 * @throws ResponseCreationException 
	 */
	private void commitChanges(Map<Integer, Metadata> changes) {
		logger.debug("Commit Changes...");
		
		// find all Subscriptions that need to be rerun and add them to
		// the changed subscriptions. We don't care about them, as they
		// are checked later on.
		List<Subscription> rerunSubs = findReRunSubscriptions(changes);
		changedSubscriptions.addAll(rerunSubs);
	
		// run all searches with the old graph for the subscriptions that need
		// to be rerun
		for (Subscription sub : rerunSubs) {
			logger.trace("Rerunning search for subscription \"" + sub.getName() + "\"");
			String sessionId = sub.getSearchRequest().getSessionId();
		
			// build up the old search result in the subscription
			sub.setReRunOldResult(new SearchResultReferenceImpl(sessionId));
			SearchResultBuilder builder = new SearchResultBuilderImpl(sessionId, sub.getReRunOldResult());
			Searcher searcher = new Searcher(identifierRep, sub.getSearchRequest(), builder);
			searcher.runSearch();
			
			//for (MetadataContainer mc : sub.getReRunOldResult().getContainers()) {
				//mc.removeSubscription(sub);
			//}
			clearContainers(sub);
			
			// Create first identifier???
			// MetadataContainer mc = identifierRep.getOrCreateIdentifier(sub.getSearchRequest().getStartIdentifier());
			// mc.addSubscription(sub);
			// sub.addContainer(mc);
		}

		// go through all changed (new or deleted) metadata
		for (Metadata m : changes.values()) {
			
			if (m.isNew() || m.isDeleted()) {
				MetadataContainer mc = m.getContainerReference();
				List<Subscription> subList = mc.getSubscriptions();
				if (mc instanceof Identifier) {
					for (Subscription sub : subList) {
						if (m.matchesFilter(sub.getSearchRequest().getResultFilter())) {
							if (m.isNew()) {
								sub.getUpdateResult().addMetadata(mc , m);
								sub.setChanged();
								changedSubscriptions.add(sub);
							} else if (m.isDeleted()) {
								sub.getDeleteResult().addMetadata(mc, m);
								sub.setChanged();
								changedSubscriptions.add(sub);
							}
						}
					}
				} else if (mc instanceof Link) {
					for (Subscription sub : subList) {
						if (m.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())
								&& m.matchesFilter(sub.getSearchRequest().getResultFilter())) {
							if (m.isNew()) {
								sub.getUpdateResult().addMetadata(mc , m);
								sub.setChanged();
								changedSubscriptions.add(sub);
							} else if (m.isDeleted()) {
								sub.getDeleteResult().addMetadata(mc, m);
								sub.setChanged();
								changedSubscriptions.add(sub);
							}
						}
					}
				}
			}
			
			// new metadata is set to state UNCHANGED
			// deleted metadata is removed from the graph
			if (m.isNew()) {
				m.setState(MetaState.UNCHANGED);
			} else if (m.isDeleted()) {
				metaController.removeReferencesTo(m);
			}
		}
	
		// HOOK IN RERUN SUBSCRIPTIONS!!!
		// Run all searches with the new graph
		// Compare the old and new SearchResult
		for (Subscription sub : rerunSubs) {
			String sessionId = sub.getSearchRequest().getSessionId();
			sub.setReRunNewResult(new SearchResultReferenceImpl(sessionId));
			SearchResultBuilder builder = new SearchResultBuilderImpl(sessionId, sub.getReRunNewResult());
			Searcher searcher = new Searcher(identifierRep, sub.getSearchRequest(), builder);
			searcher.runSearch();
			
			for (MetadataContainer mc : sub.getReRunNewResult().getContainers()) {
				mc.addSubscription(sub);
				sub.addContainer(mc);
			}
			
			compareAndGenerateNewResults(sub);
		}
	}

	/**
	 * Helper method which compares the Old and New SearchResult and appends
	 * the differences into the Delete and Update Result.
	 * 
	 * {@link Subscription} objects contain new and old {@link SearchResult}
	 * objects which are set if the {@link Subscription} is rerun.
	 * Those are of the type {@link SearchResultImplReferences}, so we can
	 * compare with ==, as we compare references.
	 * 
	 * @param sub
	 */
	private void compareAndGenerateNewResults(Subscription sub) {
		SearchResult newRes = sub.getReRunNewResult();
		SearchResult oldRes = sub.getReRunOldResult();
		
		if (newRes == null || oldRes == null) {
			logger.fatal("Whut?, newRes and oldRes have to be set!");
		}
		
		List<MetadataContainer> oldContainers = oldRes.getContainers();
		List<MetadataContainer> newContainers = newRes.getContainers();
		
		for (MetadataContainer oldContainer : oldContainers) {
			if (newContainers.contains(oldContainer)) {
				List<Metadata> oldMeta = oldRes.getMetadataFor(oldContainer);
				List<Metadata> newMeta = newRes.getMetadataFor(oldContainer);	// old == new
				
				// compare metadata
				for (Metadata m : oldMeta) {
					if (newMeta.contains(m)) {
						newMeta.remove(m);
					} else {
						// newMeta does not contain oldMetadata object -> deleteResult
						sub.getDeleteResult().addMetadata(oldContainer, m);
					}
				}
				// only those that are not in oldMeta are left in newMeta,
				// we add them to the update result of the subscription.
				for (Metadata m : newMeta) {
					sub.getUpdateResult().addMetadata(oldContainer, m);
				}
				
				// remove the oldContainer from newContainers
				// newContainers will contain only those containers that are
				// new?
				newContainers.remove(oldContainer);
			} else {
				for (Metadata m : oldRes.getMetadataFor(oldContainer)) {
					sub.getDeleteResult().addMetadata(oldContainer, m);
				}
			}
		}
	
		// only those that are not in oldContainers, but in newContainers
		for (MetadataContainer newContainer : newContainers) {
			for (Metadata m : newRes.getMetadataFor(newContainer)) {
				sub.getUpdateResult().addMetadata(newContainer, m);
			}
		}
	}

	
	
	
	/**
	 * Go through the list of {@link Metadata} objects that are NEW or DELETED,
	 * if such a objects is attached to a {@link Link} some checks have to be
	 * done to find out if the {@link Subscription} needs to be rerun.
	 * 
	 * If the {@link Metadata} object is new, and the {@link Identifier} objects
	 * of the link contain {@link Subscription} objects that the {@link Link} does
	 * not contain, and further the newly created {@link Metadata} object
	 * matches the matchlinks filter of such a {@link Subscription} the 
	 * {@link Subscription} needs to be rerun.
	 * 
	 * If a {@link Metadata} object was deleted, we need to check for every
	 * {@link Subscription} object of this {@link Link} whether another
	 * {@link Metadata} object of that {@link Link} matches the matchlinks filter
	 * of the current {@link Subscription}. If this is not the case the
	 * {@link Subscription} needs to be rerun.
	 * 
	 * @param changes the {@link Metadata} objects that have been marked as changed
	 * @return
	 */
	private List<Subscription> findReRunSubscriptions(
			Map<Integer, Metadata> changes) {
		
		HashSet<Subscription> ret = new HashSet<Subscription>();
		
		for (Metadata m : changes.values()) {
			if (m.getContainerReference() instanceof Link) {
				Link link = (Link) m.getContainerReference();
				if (m.isNew()) {
					HashSet<Subscription> set = new HashSet<Subscription>(link.getIdent1().getSubscriptions());
					set.addAll(link.getIdent2().getSubscriptions());
					
					for (Subscription sub : link.getSubscriptions()) {
						if (set.contains(sub)) {
							set.remove(sub);
						}
					}
				
					for (Subscription sub : set) {
						if (m.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())) {
							sub.setChanged();
							ret.add(sub);
						}
					}
				} else if (m.isDeleted()) {
					for (Subscription sub : link.getSubscriptions()) {
						List<Metadata> all = link.getMetadata();
						int count = 0;
						for (Metadata md : all) {
							if (!md.isDeleted() && md.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())) {
								count++;
							}
						}
						
						if (count == 0) {
							ret.add(sub);
							sub.setChanged();
						}
					}
				}
			}
		}
		return new LinkedList<Subscription>(ret);
	}

	/**
	 * If a {@link SubscriptionObserver} calls getPollResultFor() the corresponding
	 * Publisher with the given sessionId is searched.
	 * 
	 * All of it's subscriptions are checked, to see if any of them has changes.
	 * If a subscription has changes either the search result, the update result,
	 * the delete result or the notify result has elements which might be of
	 * interest for the MAPC.
	 * 
	 * A {@link PollResult} is constructed which contains {@link SearchResult}
	 * objects for all the different result types and further for all
	 * subscriptions a client has.
	 * 
	 * thoughts:
	 * 	could this get too big?
	 * 
	 * @param sessionId
	 * @return
	 * @throws NoSuchPublisherException
	 */
	public PollResult getPollResultFor(String sessionId) throws NoSuchPublisherException {
		Publisher p = publisherRep.getPublisherBySessionId(sessionId);
		Collection<Subscription> activeSubs = p.getSubscriptions().values();
		
		List<SearchResult> slist = new LinkedList<SearchResult>();
		List<SearchResult> ulist = new LinkedList<SearchResult>();
		List<SearchResult> dlist = new LinkedList<SearchResult>();
		List<SearchResult> nlist = new LinkedList<SearchResult>();
	
		
		for (Subscription sub : activeSubs) {
			if (sub.hasChanged()) {
				if (!sub.getSearchResult().isEmpty())
					slist.add(sub.getSearchResult());
				if (!sub.getUpdateResult().isEmpty()) 
					ulist.add(sub.getUpdateResult());
				if (!sub.getDeleteResult().isEmpty())
					dlist.add(sub.getDeleteResult());
				if (!sub.getNotifyResult().isEmpty())
					nlist.add(sub.getNotifyResult());
				
				sub.resetResults();
			}
		}
		
		p.setNotified(false);
		
		PollResult pr = new PollResult(p.getSessionId(), slist, ulist, dlist, nlist);
		
		return pr;
	}
	

	/**
	 * Set the {@link SubscriptionObserver} to the given one.
	 * 
	 * @param subObs
	 * @throws AlreadyObservedException if another {@link SubscriptionObserver} was
	 *  registered before
	 */
	public void setSubscriptionObserver(SubscriptionObserver subObs) throws AlreadyObservedException {
		if (observer != null)
			throw new AlreadyObservedException("Only one SubscriptionObserver is allowed!");
		
		observer = subObs;
	}

	
	
	
	/**
	 * Method which handles a notify message from a MAPC.
	 * 
	 * It is similar to publishUpdate in {@link PublishService}, however, somewhat
	 * simpler as we don't care about single or multivalue metadata.
	 * 
	 * First set the publisherId of the new {@link Metadata} object, then
	 * go through all subscriptions which are attached to the {@link MetadataContainer}
	 * and see if one of them matches that {@link Metadata} object.
	 * 
	 * {@link Link} and {@link Identifier} objects must be handled differently
	 * as different filters are used.
	 * 
	 * @param pub
	 * @param metadatacontainer
	 * @param mlist
	 */
	public void handleNotify(Publisher pub, MetadataContainer metadatacontainer, List<Metadata> mlist) {
		logger.trace("Handle notify from " + pub + " to " + metadatacontainer +
				" with " + mlist.size() + " metadata objects ");
		
		for (Metadata m : mlist) {
			m.setPublisherId(pub.getPublisherId());
			for (Subscription sub : metadatacontainer.getSubscriptions()) {
				if (metadatacontainer instanceof Identifier) {
					if (m.matchesFilter(sub.getSearchRequest().getResultFilter())) {
						logger.trace("Adding notify metadata (" + m.getType().getTypeString() + ") to subscription " + sub.getName());
						sub.getNotifyResult().addMetadata(metadatacontainer, m);
						sub.setChanged();
						changedSubscriptions.add(sub);
					}
				} else if (metadatacontainer instanceof Link) {
					if (m.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())) {
						if (m.matchesFilter(sub.getSearchRequest().getResultFilter())) {
							logger.trace("Adding notify metadata (" + m.getType().getTypeString() + ") to subscription " + sub.getName());
							sub.getNotifyResult().addMetadata(metadatacontainer, m);
							sub.setChanged();
							changedSubscriptions.add(sub);
						}
					}
				}
			}
		}
	}

	
	
	/**
	 * Simple helper method which calls pollResultAvailable() on the observer
	 * with the sessionId of the given publisher.
	 * 
	 * @param pub
	 */
	private void notifyChangedSubscriptions(Publisher pub) {
		if (observer != null) {
			if (pub.getSessionId() == null) {
				// internal error?
				logger.fatal("Whut? Publisher with no session has subscriptions?");
				return;
			}
			
			if (!pub.wasNotified()) {
				logger.trace(pub + " has new poll results");
				pub.setNotified(true);
				observer.pollResultAvailable(pub.getSessionId());
			} else {
				logger.trace(pub + " was already notified!");
			}
		}
	}
	
	
	/**
	 * Helper method to prepare the notification of the observer with all
	 * publishers that have new results available.
	 * 
	 * @param changedSubs
	 */
	private void notifyChangedSubscriptions() {
		logger.debug("Notify SubscriptionObserver about new results...");
		for (Subscription sub : changedSubscriptions) {
			try {
				Publisher pub = publisherRep.getPublisherBySessionId(sub.getSearchRequest().getSessionId());
				logger.trace("Subscription \"" + sub.getName() + "\" of " + pub + " has changes!");
				notifyChangedSubscriptions(pub);
			} catch (NoSuchPublisherException e) {
				// this is an internal error...
				logger.fatal("Whut? No Publisher available, but a subscription for him?");
			}
		}
		changedSubscriptions.clear();
	}

	/**
	 * Go through all {@link MetadataContainer} objects that this subscription
	 * is attached to and remove the reference to the given {@link Subscription}
	 * object.
	 * 
	 * @param sub
	 */
	void clearContainers(Subscription sub) {
		for (MetadataContainer mc : sub.getContainers()) {
//			logger.trace("Subs before:");
//			for (Subscription asub : mc.getSubscriptions()) {
//				logger.trace("\t" + asub.getName());
//			}
			logger.trace("Removing subscription \"" + sub.getName() + "\" from " + mc);
			mc.removeSubscription(sub);
//			logger.trace("Subs after:");
//			for (Subscription asub : mc.getSubscriptions()) {
//				logger.trace("\t" + asub.getName());
//			}
		}
		sub.getContainers().clear();
	}
}
