/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph
 * File:    Identity.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph;

import de.fhhannover.inform.iron.mapserver.configuration.datamodel.ConfigurationManager;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;

/**
 * Class which represents a Identity.
 * 
 * A Identity can be of different types which are represented
 * by a DeviceType.
 * 
 * 
 * @author aw
 * 
 * @version 0.1
 *
 * created: 26.11.09 
 * changes:
 * 	26.11.09 aw - Created first version of this class
 * 	27.11.09 aw - equals() check for null, added constructor
 *  	to be used with default administrative domain
 *  28.11.09 aw - equals() use super.equals() to compare administrative
 * 		domains.
 *  09.12.09 aw - implemented copy()
 *  17.12.09 aw - added type
 *  04.02.10 aw - toString()
 *  07.02.10 aw - modified toString(), use ConfigurationManager with new
 *                type based function to determine wheter case sensitivity is
 *                needed.
 *  12.02.10 aw - modified to use new IdentityType enum
 *  
 *  
 * what about:
 *  - fixing the equals method...
 *
 */
public class Identity extends  IdentifierWithAdministrativeDomain {
	
	private String name;
	private IdentityTypeEnum type;
	private String otherTypeDefinition;
	private static ConfigurationManager configMgr;
	static {
		configMgr = ConfigurationManager.getInstance();
	}
	
	/**
	 * Private constructor that mostly does error checking.
	 * 
	 * Checks name for null and checks length > 0
	 * 
	 * @param name
	 * @param ad
	 * @throws IdentifierConstructionException
	 */
	private Identity(String name, String ad) throws IdentifierConstructionException {
		super(IdentifierType.instanceOfTypeWith("identity"), ad);
		if (name == null || name.length() == 0)
			throw new IdentifierConstructionException("Identity: name is invalid");
		
		this.name = new String(name);
	}
	
	/**
	 * Construct a identity object from name,
	 * administrative-domain and the type.
	 * 
	 * Using this constructor it is not allowed to
	 * set it to IdentityType.other!
	 * 
	 * @param name
	 * @param ad
	 * @param it
	 * @throws IdentifierConstructionException 
	 */
	public Identity(String name, String ad, IdentityTypeEnum it) throws IdentifierConstructionException {
		this(name, ad);
		if (it == null)
			throw new IdentifierConstructionException("Identity: type is null");
		if (it == IdentityTypeEnum.other)
			throw new IdentifierConstructionException("Identity: for type other" +
					"use the other constructor!");
		type = it;
	}
	
	/**
	 * Construct a identity object from name,
	 * administrative-domain and a other-type-definition.
	 * 
	 * @param name
	 * @param ad
	 * @param otherTypeDef
	 * @throws IdentifierConstructionException
	 */
	public Identity(String name, String ad, String otherTypeDef) throws IdentifierConstructionException {
		this(name, ad);
		if (otherTypeDef == null)
			throw new IdentifierConstructionException("otherTypeDef " +
					"is not allowed to be null!");
		this.otherTypeDefinition = new String(otherTypeDef);
		this.type = IdentityTypeEnum.other;
	}
	
	/**
	 * Constructor for Identity Identifier to create an empty
	 * administrativeDomain with a type other than other.
	 * 
	 * @param name
	 * @param it
	 * @throws IdentifierConstructionException
	 */
	public Identity(String name, IdentityTypeEnum it) throws IdentifierConstructionException {
		this(name, "", it);
	}
	
	
	public int hashCode() {
		String hashString;
		hashString = super.hashCode() + type.toString();
		if (type == IdentityTypeEnum.other)
			hashString += otherTypeDefinition;
		hashString += name;
		return hashString.hashCode();
	}
	
	public IdentityTypeEnum getIdentityType() {
		return type;
	}
	
	public String getName() {
		return new String(name);
	}
	
	/**
	 * Might return null if the type of this Identity is not "other"
	 * 
	 * @return
	 */
	public String getOtherTypeDefinition() {
		if (type != IdentityTypeEnum.other)
			return null;
		
		return new String(otherTypeDefinition);
	}

	/**
	 * handle every case
	 *  - same reference
	 *  - same type is needed!
	 *  
	 *  FIXME: THIS IS COMPLETLY UGLY
	 */
	@Override
	public boolean equals(Object o) {
		boolean caseSensitive = false;
		boolean nameRes = false;
		Identity i;
		IdentityTypeEnum it;
		String iname;
		String iotherTypeDef;
		
		if (this == o)
			return true;
		
		if (super.equals(o)) {
			if (o instanceof Identity) {
				i = (Identity) o;
				it = i.getIdentityType();
				iname = i.getName();
			
				// same type ?
				if (type != it)
					return false;
			
				// request whether this type has to be checked case sensitive
				caseSensitive = configMgr.isIdentityTypeCaseSensitive(type);
				
				// compare other-type-definitions
				if (type == IdentityTypeEnum.other) {
					iotherTypeDef = i.getOtherTypeDefinition();
					
					// we check the other type def case sensitive... ?
					// FIXME: What is the specification saying?
					if (!otherTypeDefinition.equals(iotherTypeDef))		
						return false;
				}
				if (caseSensitive) {
					nameRes = name.equals(iname);
				} else {
					nameRes = name.equalsIgnoreCase(iname);
				}
			}
		}
		return nameRes;
	}

	@Override
	public Identifier copy() {
		Identity id = null;
		
		try {
			id = new Identity(this.getName(), this.getAdministrativeDomain(),
															this.getIdentityType());
		} catch (IdentifierConstructionException e) {
			e.printStackTrace();						/* should not fail */
		}
		
		return id;
	}
	
	public String toString() {
		String ad = getAdministrativeDomain();
		StringBuffer sb = new StringBuffer();
		sb.append("Identity<name=" + name);
		if (!(ad == null | ad.length() == 0)) {
			sb.append(" ad=" + ad);
		}
		sb.append(" type=" + type.toString());
		if (type == IdentityTypeEnum.other) {
			sb.append(" def=" + otherTypeDefinition);
		}
		sb.append(">");
		return sb.toString();
	}
}