/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph
 * File:    MacAddress.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;
import de.fhhannover.inform.iron.mapserver.utils.MacAddressValidator;

/**
 * MACAddress Type
 * 
 * This class represents a MAC Address Identifier
 * A MAC address must consist of 6 times 2 hexdigits in lowercase!
 * 
 * @author aw
 * @version 0.1
 * 
 * created: 26.11.09
 * changes: 
 *  26.11.09 aw - Created first version of this class
 *  27.11.09 aw - Implemented equals() and some logic
 *	28.11.09 aw - equals() use super.equals() to compare administrative
 *		domains.
 *	09.12.09 aw - implemented copy()
 *	17.12.09 aw - added type
 *  04.02.10 aw - toString()
 *  07.02.10 aw - modified toString()
 *  08.02.10 aw - added hashCode(), doh!
 */
public class MacAddress extends IdentifierWithAdministrativeDomain {
	
	private String value;
	
	/**
	 * Create a MAC Address based on value and administrative-domain
	 * 
	 * @param value
	 * @param ad
	 * @throws IdentifierConstructionException
	 */
	public MacAddress(String value, String ad) throws IdentifierConstructionException {
		super(IdentifierType.instanceOfTypeWith("mac-address"), ad);
		if (value == null)
			throw new IdentifierConstructionException("MacAddress: value null");
		
		if (!MacAddressValidator.validateMacAddress(value)) {
			throw new IdentifierConstructionException("MacAddress: not a valid" +
					"mac address");
		}
		this.value = new String(value);
	}
	
	/**
	 * Create a MAC address based only on the value.
	 * The administrativeDomain is set to ""
	 * 
	 * @param value
	 * @throws IdentifierConstructionException
	 */
	public MacAddress(String value) throws IdentifierConstructionException {
		this(value, "");
	}
	
	public String getValue() {
		return new String(value);
	}
	
	public int hashCode() {
		String res = super.hashCode() + value + "";
		return res.hashCode();
	}

	@Override
	public boolean equals(Object o) {
		boolean valRes = false;
		MacAddress omac;
		
		if (o == null)
			return false;
		
		if (this == o)
			return true;
		if (super.equals(o)) {
			if (o instanceof MacAddress) {
				omac = (MacAddress) o;
				valRes = value.equals(omac.getValue());
			}
		}
		return valRes;
	}

	@Override
	public Identifier copy() {
		MacAddress mac = null;
		try {
			mac = new MacAddress(this.getValue(), this.getAdministrativeDomain());
		} catch (IdentifierConstructionException e) {
			e.printStackTrace();					/* should not fail */
		}
		return mac;
	}

	@Override
	public String toString() {
		String ad = getAdministrativeDomain();
		StringBuffer sb = new StringBuffer();
		sb.append("MacAddress<value=" + value);
		if (!(ad == null | ad.length() == 0)) {
			sb.append(" ad=" + ad);
		}
		sb.append(">");
		return sb.toString();
	}
}