/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph
 * File:    MetadataContainer.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaCardinalityType;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaState;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaType;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.datamodel.Filter;
import de.fhhannover.inform.iron.mapserver.datamodel.Subscription;
import de.fhhannover.inform.iron.mapserver.logging.Logging;

/**
 * Represents the superclass of {@link Identifier} and {@link Link}.
 * It combines functionality and represents the most abstract element
 * which is contained in the graph
 * 
 * Adding and removing metadata from identifiers and links is done
 * identically as well as searching on those.
 * 
 * FIXME:
 * Use ArrayList or LinkedList?
 * 
 * FIXME:
 * Subscriptions and multiValue vs singleValue handling (removing multiValue md,
 * when someone does publish singelValue should be seen in subscriptions???)
 *
 * @author awe, vp
 * @version 0.1
 * created: 19.11.09
 *
 *
 */

/*
 * changes:
 * 	01.12.09 vp - empty methods implemented
 * 	05.12.09 aw - implemented logig for addMetaData() and removeMetaData()
 *  10.12.09 vp - removeMetadata(Filter f) implemented 
 *  10.12.09 aw - removeMetadata needs to notify the publisher about the
 *  	removed metadata
 *  10.12.09 vp - removeMetadata remove metadata from publisher
 *  22.12.09 vp - remove Deprecated by getMetadata()
 *  04.02.09 aw - getMetadata() is still there?
 *                added logging with log4j.
 *  07.02.10 aw - made this class abstract, added toString()
 *  08.02.10 aw - removeMetadata() changed so that no call is made to the
 *                original publisher,
 *                removeMetadata(Filter) removed
 *  10.02.10 aw - removed configureLogging()
 *  
 *  12.04.10 aw - change removeMetadata() to accept a List of removed
 *                metadata. This is needed in upper layers to remove
 *                it from the responsible publisher
 *  03.05.10 aw, vp - use metastate
 *  18.06.10 aw - added getMetadataInGraph() methods to be used, if only those
 *   metadata objects are wanted for which isInGraph() returns true;
 *  
 */
abstract public class MetadataContainer {
	
	/**
	 * One static logger instance used by everybody
	 */
	protected static Logger logger;
	static {
		logger = Logging.getTheLogger();
	}

	/**
	 * List of metadata objects
	 */
	private List<Metadata> metaDataList;
	
	/**
	 * List of subscriptions this metadatacontainer is in
	 */
	private List<Subscription> subscriptions;
	
	
	
	/**
	 * DefaulConstructor
	 */
	public MetadataContainer() {
		metaDataList = new LinkedList<Metadata>();
		subscriptions = new LinkedList<Subscription>();
	}
 
	/**
	 * First make a deep copy of the incoming metadata.
	 * Add it depending whether it is single or multivalue.
	 * 
	 * multiValue metadata is simply appended to the list of metadata objects.
	 * singleValue metadata needs to replace the old instance of that metadata
	 * object.
	 * 
	 * The implementation of addSingleValueMetadata() and addMultiValueMetadata()
	 * check for existing multiValue and singleValue metadata.
	 * 
	 * 
	 * @param m Metadata to add
	 * @param rm Metadata which was removed
	 * @return a reference to the Metadata Object that was really inserted,
	 * 		   this is needed for the PublisherRepository.
	 */
	public Metadata addMetadata(Metadata m, List<Metadata> rm) {
		assert (rm != null);
		if (m != null) {
			m = m.copy();
			if (m.getCardinality() == MetaCardinalityType.multi) {
				addMultiValueMetadata(m, rm);
			} else if (m.getCardinality() == MetaCardinalityType.single) {
				addSingleValueMetadata(m, rm);
			}
			m.setState(MetaState.NEW);
		}
		return m;
	}
	
	/**
	 * Just add the given Metadata as it is multiValue.
	 * But we check if there is any Metadata of this type
	 * that has a singleValue cardinality.
	 * 
	 * @param m
	 */
	private void addMultiValueMetadata(Metadata m, List<Metadata> rm) {
		if (m == null) {
			return;
		}
		
		List<Metadata> sameTypeMetadata = searchForType(m.getType());
		
		for (Metadata md : sameTypeMetadata) {
			if (md.getCardinality() == MetaCardinalityType.single) {
				logger.warn("Found singleValue Metadata of same type where it" +
				"should be multiValue. Removing it");
				removeMetadata(md);
				rm.add(md);
			}
		}
		metaDataList.add(m);
		logger.trace("Metadatalist of " + this + " size=" + metaDataList.size());
	}
	
	/**
	 * Replace Metadata that is already there
	 * Check if metadata that needs to be removed are of same
	 * cardinality. If not output error. Additionally output
	 * an error if we removed more than one metadata.
	 * 
	 * @param m
	 */
	private void addSingleValueMetadata(Metadata m, List<Metadata> rm) {
		if (m == null)
			return;
		
		List<Metadata> sameTypeMetadata = searchForType(m.getType());
		
		for (Metadata md : sameTypeMetadata) {
			if (md.getCardinality() == MetaCardinalityType.multi) {
				logger.warn("Found multiValue Metadata of same type where it" +
										"should be singleValue. Removing it");
			}
			removeMetadata(md);
			rm.add(md);
		}
		if (rm.size() > 1) {
			logger.warn("Removed more than one metadata while adding singleValue" +
																	"Metadata");
		}
		metaDataList.add(m);
	}
	
	/**
	 * Return a list of metadata objects (references to the real
	 * metadata objects) for which have the specified MetaDataType.
	 * 
	 * @param type
	 * @return
	 */
	private List<Metadata> searchForType(MetaType type) {
		List<Metadata> res = new LinkedList<Metadata>();
		for (Metadata md : metaDataList) {
			if (md.getType() == type) {
				res.add(md);
			}
		}
		return res;
	}
	
	/**
	 * Remove a given metadata object from the metadatalist.
	 * 
	 * @param m
	 * @return true if element was removed, false if not
	 */
	public boolean removeMetadata(Metadata m) {
		if (m != null) {
			return metaDataList.remove(m);
		}
		return false;
	}
	 
	/**
	 * Return a new List of {@link Metadata} objects associated
	 * with this {@link MetadataContainer} object.
	 * 
	 * @return list containing the metadata
	 */
	public List<Metadata> getMetadata() {
		return new ArrayList<Metadata>(metaDataList);
	}
	
	/**
	 * Return a list of {@link Metadata} objects which are associated
	 * with this {@link MetadataContainer}, however only those where
	 * isInGraph() is true, therefore only {@link Metadata} objects
	 * with state UNCHANGED
	 * 
	 * @return
	 */
	public List<Metadata> getMetadataInGraph() {
		List<Metadata> res = new ArrayList<Metadata>();
		for (Metadata m : metaDataList)
			if (m.isInGraph())
				res.add(m);
			
		return res;
	}
	 
	/**
	 * Return a list of {@link Metadata} objects, which are attached
	 * to this {@link MetadataContainer} and match the given {@link Filter}
	 * object.
	 * 
	 * @param f
	 * @return
	 */
	public List<Metadata> getMetadata(Filter f) {
		List<Metadata> res = new ArrayList<Metadata>();
		for (Metadata m : metaDataList) {
			if (m.matchesFilter(f))
				res.add(m);
		}
		return res;
	}

	/**
	 * Return a list of {@link Metadata} objects, which are attached
	 * to this {@link MetadataContainer} and match the given {@link Filter}
	 * object, however only those where isInGraph() is true, therefore only
	 * {@link Metadata} objects with state UNCHANGED
	 * 
	 * 
	 * @param f
	 * @return
	 */
	public List<Metadata> getMetadataInGraph(Filter f) {
		List<Metadata> res = new ArrayList<Metadata>();
		for (Metadata m : metaDataList) {
			if (m.matchesFilter(f) && m.isInGraph())
				res.add(m);
		}
		return res;
	}

	
	abstract public String toString();

	public List<Subscription> getSubscriptions() {
		return subscriptions;
	}

	public void removeSubscription(Subscription sub) {
		subscriptions.remove(sub);
	}
	
	public void addSubscription(Subscription sub) {
		subscriptions.add(sub);
	}
}
 
