/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph.meta
 * File:    Metadata.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph.meta;


import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import de.fhhannover.inform.iron.mapserver.datamodel.Filter;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.MetadataConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.clientmapping.Publisher;
import de.fhhannover.inform.iron.mapserver.logging.Logging;

/**
 * Abstract Metadata class.
 * 
 * Concrete {@link Metadata} classes have to implement the abstract methods
 * given in this class.
 * 
 * @author aw
 * @version 0.1
 */

/*
 * created: 28.11.09
 * changes:
 * 	28.11.09 aw - Created class and modified it to be abstract
 *  05.12.09 aw - Added MetaDataType, getSetCardinality() as abstract method,
 *  	as the cardinality is stored in the metadata itself.
 *  10.12.09 aw - Added (set|get)PublisherReference() and reference to
 *  	know which publisher published this metadata.
 *  04.02.10 aw - logging with log4j
 *  10.02.10 aw - removed configureLogging(),
 *                some renaming and cleanup
 *  11.02.10 aw,dl - added toXMLString() as abstract method
 *  18.02.10 aw - removed this method... it was just a hack
 *                to get xml conversion to work!
 *  10.04.10 aw - added abstract toW3CDocument() this is needed
 *                in the api layer and probably makes sense to
 *                have it here instead of doing conversions do
 *                W3C documents with instanceof checks
 *                
 *                lifetime is known without looking at the metadata
 *                use it in constructor
 *                
 *                Some comments
 *  11.04.10 aw - give the timestamp to the constructor, add a getTimestamp()
 *  		      method for the new timestamp field.
 *  03.05.10 aw, vp - add metastate
 *  13.06.10 aw - removed inGraph variable, renamed isDelete() to isDeleted()
 */

abstract public class Metadata {
 
	public static final String CARDINALITYSTRING = "ifmap-cardinality";
	public static final String CARDINALITYSINGLE = "singleValue";
	public static final String CARDINALITYMULTI = "multiValue";
	 
	/**
	 * a static logger
	 */
	protected final static Logger logger;
	static {
		logger = Logging.getTheLogger();
	}
	
	/**
	 * Cardinality 
	 */
	protected MetaCardinalityType cardinality = null;
	
	/**
	 * Used for comparisons between Metadata types
	 */
	private MetaType type;
	 
	private MetaLifeTimeType lifeTime;
	 
	/**
	 * Reference to the MetadataContainer where this
	 * metadata object is attached to.
	 */
	private MetadataContainer metadataContainer;
	
	/**
	 * A reference to the Publisher that published this {@link Metadata}
	 * object.
	 */
	private Publisher publisher = null;
	
	
	/**
	 * Timestamp in form of a string.
	 */
	private String timestamp;
	
	
	/**
	 * The state in which this {@link Metadata} object is.
	 */
	private MetaState state;
	
	/**
	 * Constructor which enforces subclasses to set the type;
	 * 
	 * @param type
	 * @throws MetadataConstructionException 
	 */
	protected Metadata(MetaType type, MetaLifeTimeType lifetime, String timestamp)
											throws MetadataConstructionException {
		
		if (type == null) 
			throw new MetadataConstructionException("type is not allowed to" + 
																	"be null");
		if (lifetime == null)
			throw new MetadataConstructionException("lifetime is not allowed to" + 
																	"be null");
		
		if (timestamp == null || timestamp.length() == 0)
			throw new MetadataConstructionException("timestamp has to be set");
			
		this.type		= type;
		this.lifeTime	= lifetime;
		this.timestamp	= new String(timestamp);
	}
	
	/**
	 * Copy constructor.
	 * Copy cardinality, type and lifeTime as well
	 * as the reference to the metadatas container.
	 * 
	 * @param m
	 * @throws MetadataConstructionException 
	 */
	protected Metadata(Metadata m) throws MetadataConstructionException {
		this(m.getType(), m.getLifeTime(), m.getTimestamp());
		cardinality = m.getCardinality();
		metadataContainer = m.getContainerReference();
	}

	/**
	 * Returns the timestamp that was given during construction
	 * of the object.
	 * @return
	 */
	protected String getTimestamp() {
		if (timestamp == null)
			throw new RuntimeException("timestamp not set, fix that!");
		return new String(timestamp);
	}

	/**
	 * This method should contain the logic to check
	 * whether or not this {@link Metadata} object matches the
	 * given {@link Filter} f.
	 * 
	 * @param f
	 * @return
	 */
	abstract public boolean matchesFilter(Filter f);
	 
	/**
	 * {@link Metadata} from a MAPC comes without the publisherID
	 * this must be added here to occur in search- and
	 * subscription results.
	 * 
	 * @param pubid
	 */
	abstract public void setPublisherId(String pubid);
	
	
	/**
	 * Conversation from the internal representation to a
	 * W3C document representation. This is needed when
	 * transforming this {@link Metadata} object into objects
	 * which will later be marshaled and sent to a MAPC.
	 * 
	 * @return a W3C document representation of the metadata object
	 */
	abstract public Document toW3CDocument();
	 
	/**
	 * Metadata from a MAPC comes without a timestamp. For a timestamp to
	 * occur in search and subscription results. a concrete implementation
	 * needs to be set.
	 * 
	 * This method should be called somewhere at the end of an implementations
	 * constructor.
	 * 
	 * If the timestamp can't be set the method should throw a
	 * MetadataConstructionException.
	 * 
	 * @param ts
	 */
	abstract public void setTimestamp() throws MetadataConstructionException;
	
	/**
	 * This method should set the cardinality attribute with singleValue
	 * or multiValue depending on the actual metadata during construction
	 * of an concrete implementations object.
	 * 
	 * This method should be called somewhere at the end of an implementations
	 * constructor.
	 * 
	 * If the cardinality can't be set the method should throw a
	 * MetadataConstructionException.
	 * 
	 * @throws MetadataConstructionException
	 */
	abstract public void setGetCardinality() throws MetadataConstructionException;
	 
	/**
	 * Get a reference to the MetaDataType. This can be used to compare
	 * two metadata objects to be from the same metadata type.
	 * 
	 * For example two metadata objects as <meta:role> have the same
	 * MetaDataType reference.
	 * 
	 * @return
	 */
	public MetaType getType() {
		return type;
	}
	
	/**
	 * Get the cardinality. Be aware, if the cardinality wasn't initialized,
	 * this will throw a RuntimeException!
	 * 
	 * @return
	 */
	public MetaCardinalityType getCardinality() {
		if (cardinality == null) {
			throw new RuntimeException("cardinality was not initialized, fix that!");
		}
		return cardinality;
	}
	 
	/**
	 * Return the reference to the Contanier where the
	 * Metadata object is attached
	 * 
	 * @return
	 */
	public final MetadataContainer getContainerReference() {
		return metadataContainer;
	}
	
	/**
	 * Set the Container where the Metadata object is
	 * attached.
	 * 
	 * @param m
	 */
	public final void setContainerReference(MetadataContainer m) {
		if (m != null) {
			if (metadataContainer != null)
				throw new RuntimeException("metadataContainer field was already set, fix that!");
				
			metadataContainer = m;
		}
	}

	/**
	 * Create a copy of the metadata object.
	 * The metadata should be copied, for the
	 * rest Metadata(Metadata m) could be used.
	 * 
	 * @return
	 */
	abstract public Metadata copy();
	
	
	public final MetaLifeTimeType getLifeTime() {
		if (lifeTime == null)
			throw new RuntimeException("lifeTime field not set, fix that!");
		return lifeTime;
	}
	
	/**
	 * This method is used to set a reference to the
	 * publisher who published the metadata and
	 * should only be called once. A second call
	 * will bite you.
	 * 
	 * @param p
	 */
	public void setPublisherReference(Publisher p) {
		if (p == null)
			throw new RuntimeException("tried to setting publisher to null, fix that!");
			
		if (publisher != null)
			throw new RuntimeException("publisher field was already set, fix that!");
			
		publisher = p;
	}
	
	public Publisher getPublisherReference() {
		if (publisher == null)
			throw new RuntimeException("publisher field not set, fix that!");
		return publisher;
	}

	public void setState(MetaState newState) {
		if (newState == null)
			throw new RuntimeException("tried to set state to null, fix that!");
		state = newState;
	}
	
	public MetaState getState() {
		if (state == null)
			throw new RuntimeException("state field not set, fix that!");
		return state;
	}

	public boolean isNew() {
		return state == MetaState.NEW;
	}
	
	public boolean isDeleted() {
		return state == MetaState.DELETED;
	}

	/**
	 * Only unchanged and deleted are "in the graph" NEW is not in the graph
	 * and DELETE will not be there if really DELETED.
	 * 
	 * @return
	 */
	public boolean isInGraph() {
		return (state == MetaState.UNCHANGED) || (state == MetaState.DELETED);
	}
}
 
