/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.utils
 * File:    MetadataTransformer.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.utils;

import java.util.ArrayList;
import java.util.List;

import org.apache.xerces.dom.ElementNSImpl;
import org.jdom.Document;
import org.jdom.input.DOMBuilder;
import org.trustedcomputinggroup._2010.ifmap._2.MetadataListType;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.MetadataConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaLifeTimeType;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.XMLMetadata;

/**
 * Class which contains methods to transform metadata
 * 
 * @author aw, dl
 * @version 0.1
 */

/*
 * created: 19.02.10
 * changes:
 * 19.02.10 aw - Moved methods from MessageTransformer and renamed them,
 *               commented out syserrs
 * 10.04.10 aw - use new toW3CDocument() method instead of instanceof checks
 *               when transforming from internal Metadata to org.w3c.dom.Document,
 *               
 *               use new XMLMetadata constructor which needs the MetaLifeTime
 * 11.04.10 aw - add parameters for the timestamp
 *
 */
public class MetadataTransformer {
	
	private static MetadataTransformer instance;
	
	private MetadataTransformer() { }

	public static MetadataTransformer getInstance() {
		if (instance == null)
			instance = new MetadataTransformer();
		return instance;
	}
	
	/**
	 * Method to transform the autogenerated MetadataList into a
	 * List of datamodel Metadata. 
	 * 
	 * This method transforms to the XMLMetadata implementation
	 * using DOMBuilder from JDOM to create a JDOM Document from
	 * the Document objects in the MetadataListType.
	 * 
	 * 
	 * @param metaDataListType
	 * @return List of Metadata that the datamodel can work with
	 * @throws MetadataConstructionException
	 */
	public List<Metadata> transformMetadata(MetadataListType metaDataListType,
			MetaLifeTimeType lifetime, String timestamp) throws MetadataConstructionException {
		List<Metadata> lmeta = new ArrayList<Metadata>();
	//	System.err.println("=== MetadataList ===");
		List<Object> lo = metaDataListType.getAny();
		for (Object o : lo) {
			
			org.w3c.dom.Document document = ((ElementNSImpl)o).getOwnerDocument();
			DOMBuilder builder = new DOMBuilder();
			Document jdom_document = builder.build(document);

			XMLMetadata xmlmeta;
			try {
				xmlmeta = new XMLMetadata(jdom_document, lifetime, timestamp);
			} catch (Exception e) {
				// throw up anything as a MetaDataConstructionException
				throw new MetadataConstructionException(e.getMessage());
			}
			lmeta.add(xmlmeta);
			
	//		System.err.println("=========================================");
		}
		return lmeta;
	}
	
	
	/**
	 * This method converts from a datamodel metadata list to a
	 * MetadataListType.
	 * It simply uses the toW3CDocument() method of Metadata implementations
	 * and puts those into the MetaDataListType
	 * 
	 * Warning, returns null if no metadata is available in the 
	 * given list.
	 * 
	 * @param List of metadata from the datamodel
	 * @return Type which can be set in the autogenerated classes as
	 *         list
	 */
	public MetadataListType transformMetadata(List<Metadata> metadata) {
		MetadataListType ret = null;
		if (metadata != null && metadata.size() > 0) {
			ret = new MetadataListType();
			
			for (Metadata m : metadata) {
				org.w3c.dom.Document w3cDoc =  m.toW3CDocument();
				if (w3cDoc != null) {
					ret.getAny().add(w3cDoc.getDocumentElement());
				}
			}
		}
		return ret;
	}

}
